from pymemdb import Table, ColumnDoesNotExist


def test_add_columns():
    t = Table(primary_id="id")
    t.create_column("a")
    t.create_column("b")

    assert t.columns == ["id", "a", "b"]


def test_delete_columns():
    t = Table(primary_id="primary_key")
    t.create_column("a")
    t.create_column("b")
    del t["b"]

    assert t.columns == ["primary_key", "a"]


def test_len_of_column():
    t = Table()
    t.insert({"a": 1})
    t.insert({"a": 1})
    t.insert({"a": 1})

    assert len(t["a"]) == 3




import pytest

from pymemdb import Table

@pytest.fixture(scope="session")
def big_table():
    t = Table(primary_id="normal")
    for i in range(100):
        t.insert({"normal": i, "squared": i**2, "cubed": i**3})
    return t


import pytest

from pymemdb import Table, UniqueConstraintError, ColumnDoesNotExist

def test_unique_constraint():
    t = Table()
    t.create_column("a", unique=True)
    t.insert({"a": 1, "b": 2})
    t.insert({"a": 2})

    with pytest.raises(UniqueConstraintError):
        t.insert({"a": 1})


def test_find_col_not_exists(big_table):
    with pytest.raises(KeyError):
        result = list(big_table.find(quadratic=17, ignore_errors=False))

    result = list(big_table.find(quadratic=17, ignore_errors=True))
    assert result == []


def test_delete_invalid_columns():
    t = Table()
    t.create_column("a")
    t.create_column("b")

    with pytest.raises(ColumnDoesNotExist):
        del t["c"]


def test_invalid_delete():
    t = Table(primary_id="a")
    row = {"a": 1, "foo": "bar"}
    t.insert(row)
    t.delete(a=5, ignore_errors=True)
    assert list(t.all()) == [row]

    with pytest.raises(KeyError):
        t.delete(a=5, ignore_errors=False)

    assert t.delete(a=1, ignore_errors=False) == 1
    assert len(t) == 0


def test_invalid_colname():
    t1 = Table()
    with pytest.raises(ColumnDoesNotExist):
        t1["g"]


def test_insert_pk_twice():
    t = Table(primary_id="pk")
    t.insert(dict(pk=1, b=2))
    with pytest.raises(UniqueConstraintError):
        t.insert(dict(pk=1, b=3))


def test_invalid_arg_ordered():
    t = Table(primary_id="a")
    t.insert({"a": 1})

    with pytest.raises(ValueError):
        list(t.all(ordered="foo"))



from typing import Hashable, Set

from collections import defaultdict
from .errors import UniqueConstraintError


class Column:

    def __init__(self, default: Hashable = None, unique: bool = False):
        self.cells: dict = dict()
        self.values: defaultdict = defaultdict(set)
        self.default = default
        self.unique = unique

    def insert(self, pk: int, val: Hashable) -> None:
        if self.unique and val in self.values:
            raise UniqueConstraintError(f"{val} already present in column "
                                        f"(row {self.values[val]})")
        self.cells[pk] = val
        self.values[val].add(pk)

    def drop(self, pk: int) -> None:
        if pk in self.cells:
            val = self.cells[pk]
            del self.cells[pk]
            self.values[val].remove(pk)
            if not self.values[val]:
                del self.values[val]

    def find(self, val: Hashable) -> Set:
        return self.values.get(val, set())

    def find_value(self, pk: int) -> Hashable:
        return self.cells.get(pk, self.default)

    def __len__(self):
        return len(self.cells)



from __future__ import annotations
from typing import List, Optional

import dataset

from pymemdb import TableAlreadyExists
from pymemdb import Table


class Database:

    def __init__(self) -> None:
        self._tables: dict = dict()

    def create_table(self, name: str, primary_id: str = "id") -> Table:
        if name in self._tables:
            raise TableAlreadyExists(name)

        self._tables[name] = Table(name, primary_id=primary_id)

        return self._tables[name]

    def drop_table(self, name: str) -> None:
        self._tables[name].drop()
        del self._tables[name]

    def __getitem__(self, name: str):
        if name not in self._tables:
            table = self.create_table(name)
            return table
        return self._tables[name]

    def __setitem__(self, key, item):
        if key in self._tables:
            raise TableAlreadyExists(key)
        if not isinstance(item, Table):
            raise TypeError(f"{item} not an instance of 'Table'!")
        self._tables[key] = item

    @property
    def tables(self) -> List[Optional[str]]:
        return list(self._tables)

    def to_dataset(self, db: dataset.Database, chunk_size: int = 1000) -> dataset.Database:
        for tablename in self.tables:
            pk_id = self._tables[tablename].idx_name
            tableobject = db.create_table(tablename, primary_id=pk_id)
            tableobject.insert_many(list(self._tables[tablename].all()),
                                    chunk_size=chunk_size)

        return db

    @classmethod
    def from_dataset(cls, db: dataset.Database) -> Database:
        pymemdb_database = cls()

        for tablename in db.tables:
            pymemdb_database[tablename] = Table.from_dataset(db[tablename])

        return pymemdb_database



from pymemdb import Table, UniqueConstraintError, ColumnDoesNotExist


def test_number_of_rows():
    table = Table(primary_id="pk")
    rows = []
    for i in range(167):
        row = {"pk": i, "a": 1, "b": 2, "c": 0}
        table.insert(row)
        rows.append(row)

    assert list(table.all()) == rows


def test_delete_simple():
    t = Table()
    t.insert(dict(b=2))
    t.insert(dict(a="John"))
    t.insert(dict(a="John"))
    t.delete(a="John")

    assert list(t.find(a="John")) == []
    assert len(list(t.all())) == 1


def test_delete_find_delete():
    t = Table(primary_id="a")
    t.insert({"a": 1, "b": 2})
    t.insert({"a": 2, "b": 5})
    t.insert({"a": 3, "b": 6})

    result = list(t.find(a=1))
    assert len(result) == 1
    assert len(t) == 3

    n_delete = t.delete(a=1)

    result = list(t.find(a=1))
    print(result)
    assert n_delete == 1
    assert len(result) == 0
    assert len(t) == 2
    assert list(t.all(ordered="ascending")) == [{"a": 2, "b": 5},
                                                {"a": 3, "b": 6}]


def test_update():
    t = Table(primary_id="id")
    t.insert({"a": 1, "b": 2})
    t.insert({"a": 2, "b": 5})
    t.insert({"a": 3, "b": 6})
    t.insert({"a": 3, "b": 7})
    t.insert({"a": 3, "b": 8})

    t.update(where={"a": 2}, b=5)
    result = next(t.find(a=2))
    assert result["b"] == 5
    t.update(dict(a=3), b=10)

    result = list(t.find(a=3))
    assert all(row["b"] == 10 for row in result)


def test_update_no_vals():
    t = Table(primary_id="a")
    row = {"a": 1, "b": 2}
    t.insert({"a": 1, "b": 2})
    t.update(where={"a": 2}, b=5)
    assert list(t.all()) == [row]


def test_insert_ignore(big_table):
    big_table.insert(dict(normal=101, squared=102))

    assert next(big_table.find(normal=101))["squared"] == 102
    assert len(big_table) == 101

    for _ in range(3):
        big_table.insert_ignore(dict(normal=102, squared=102),
                                keys=["normal", "squared"])

    assert len(big_table) == 102
    assert len(list(big_table.find(normal=102))) == 1
    assert len(list(big_table.find(squared=102))) == 2


def test_get_all_ordered_descending():
    t = Table(primary_id="a")
    t.insert({"a": 1, "b": 2})
    t.insert({"a": 2, "b": 5})
    t.insert({"a": 3, "b": 6})

    results = t.all(ordered="descending")

    assert [r["b"] for r in results] == [6, 5, 2]


def test_equal_tables():
    t1 = Table(name="foo")
    t2 = Table(name="bar")
    t3 = Table(name="foo")

    assert t1 == t3
    assert t1 != t2




import os
import pytest

import dataset

from pymemdb import Database, Table, TableAlreadyExists


def test_database_create():
    db = Database()
    db.create_table("mytable", primary_id="test")

    assert db.tables == ["mytable"]
    assert db["mytable"].idx_name == "test"


def test_drop_table():
    db = Database()
    db.create_table("mytable", primary_id="test")
    db.create_table("mytable2", primary_id="test")
    db.create_table("mytable3", primary_id="test")

    db.drop_table("mytable2")

    assert db.tables == ["mytable", "mytable3"]


def test_delete_invalid_table():
    db = Database()

    with pytest.raises(KeyError):
        db.drop_table("foo")


def test_access_table():
    db = Database()
    assert isinstance(db["mytable"], Table)
    assert db["mytable"].name == "mytable"


def test_assign_new_table():
    t = Table()
    db = Database()
    db["test"] = t

    assert "test" in db.tables


def test_assign_wrong_type():
    db = Database()
    with pytest.raises(TypeError):
        db["mytable"] = "Hello World"


def test_error_on_overwrite():
    t1 = Table()
    t2 = Table()
    db = Database()
    db["mytable"] = t1
    exp_msg = "Table 'mytable' already exists in the Database!"
    with pytest.raises(TableAlreadyExists, match=exp_msg):
        db["mytable"] = t2


def test_sqlitedb_create(tmpdir):
    db = Database()
    db.create_table("a", primary_id="col1")
    db.create_table("b", primary_id="col1b")
    db["a"].insert({"col1": 1, "col2": "stringcol"})
    db["b"].insert({"col1b": 1, "col2b": 3.5})

    dataset_db = dataset.connect(f"sqlite:///{tmpdir / 'test.db'}")
    sqlitedb = db.to_dataset(dataset_db, chunk_size=1)

    assert list(sqlitedb["a"].all()) == list(db["a"].all())
    assert list(sqlitedb["b"].all()) == list(db["b"].all())
    assert os.path.isfile(tmpdir/"test.db")


def test_raise_table_already_exists():
    db = Database()
    db.create_table("table")

    with pytest.raises(TableAlreadyExists):
        db.create_table("table")



import pytest

import dataset

import pymemdb

@pytest.fixture(scope="function")
def simple_dataset_db():
    db = dataset.connect("sqlite:///:memory:", row_type=dict)
    simple_table = db.create_table("my_table")
    simple_table.insert_many([dict(a=i, b=1) for i in range(10)])

    return db


def test_simple_table(simple_dataset_db):
    simple_table = simple_dataset_db["my_table"]
    pytable = pymemdb.Table.from_dataset(simple_table)

    assert pytable.name == "my_table"
    assert list(pytable.all()) == list(simple_table.all())

def test_simple_table_non_default_pk():
   # simple_table = simple_dataset_db["my_table"]

    db = dataset.connect("sqlite:///:memory:", row_type=dict)
    simple_table = db.create_table("my_table", primary_id="a")
    simple_table.insert_many([dict(a=i, b=1) for i in range(10)])
    pytable = pymemdb.Table.from_dataset(simple_table)

    testrow = dict(a=21, b=2)
    simple_table.insert(testrow)
    pytable.insert(testrow)

    assert list(pytable.all()) == list(simple_table.all())


def test_to_dataset_with_drop(simple_dataset_db):
    rows = [{"b": i, "c": 3} for i in range(1, 6)]
    t = pymemdb.Table("my_table", primary_id="id")
    for r in rows:
        t.insert(r)

    t.to_dataset(simple_dataset_db, drop=True)
    assert list(simple_dataset_db["my_table"].all()) == list(t.all())

@pytest.mark.xfail
def test_to_dataset_with_drop_change_pk(simple_dataset_db):
    """known strange behavior addressd in
        https://github.com/pudo/dataset/issues/329
    """
    rows = [{"b": i, "c": 3} for i in range(1, 6)]
    t = pymemdb.Table("my_table", primary_id="b")
    for r in rows:
        t.insert(r)

    t.to_dataset(simple_dataset_db, drop=True)
    assert list(simple_dataset_db["my_table"].all()) == list(t.all())


def test_to_dataset_no_drop(simple_dataset_db):
    rows = [{"b": i, "c": 3} for i in range(1, 6)]
    t = pymemdb.Table("my_table2", primary_id="b")
    for r in rows:
        t.insert(r)

    t.to_dataset(simple_dataset_db, drop=False)

    assert list(simple_dataset_db["my_table2"].all()) == list(t.all())



from pymemdb import Table

import pytest

def test_find_rows(big_table):
    result = next(big_table.find(normal=8))
    assert result["squared"] == 64


def test_find_in(big_table):
    result = list(big_table.find(normal=[7, 8]))
    assert {row["squared"] for row in result} == {49, 64}


def test_find_multiple_keys(big_table):
    result = list(big_table.find(normal=3, squared=9))
    assert len(result) == 1
    assert result[0]["cubed"] == 27


def test_no_find_multiple_keys(big_table):
    result = list(big_table.find(normal=3, squared=10))
    assert len(result) == 0


def test_find_with_default_val():
    t = Table(primary_id="b")
    t.insert(dict(b=1))
    t.insert(dict(b=2))
    t.insert(dict(a=1, b=3))
    results = list(t.find(a=None))
    assert results == [dict(b=1, a=None), dict(b=2, a=None)]


def test_find_with_multiple_default_vals():
    t = Table(primary_id="b")
    t.create_column("firstname", default="John")
    t.create_column("lastname", default="Smith")
    t.insert(dict(b=1))
    t.insert(dict(b=2))
    t.insert(dict(b=3, lastname="Doe"))

    result = t.find(lastname="Smith")
    result2 = t.find(b=2)
    result3 = t.find(firstname="John", lastname="Smith")

    assert [r["b"] for r in result] == [1, 2]
    assert list(result2) == [dict(b=2, firstname="John", lastname="Smith")]
    assert [r["b"] for r in result3] == [1, 2]


def test_find_no_results():
    t = Table(primary_id="a")
    t.insert(dict(a=1, b=2))

    with pytest.raises(StopIteration):
        next(t.find(a=3))

    assert next(t.find(a=1)) == dict(a=1, b=2)


def test_find_one_single_result():
    t = Table(primary_id="a")
    t.insert(dict(a=1, b=2))

    row = t.find_one(a=1)

    assert row == dict(a=1, b=2)


def test_find_one_no_result():
    t = Table(primary_id="a")
    t.insert(dict(a=1, b=2))

    row = t.find_one(a=2)

    assert row == None


class TableAlreadyExists(Exception):

    def __init__(self, tablename):
        self.tablename = tablename
        super(TableAlreadyExists, self).__init__()

    def __str__(self):
        return f"Table '{self.tablename}' already exists in the Database!"


class ColumnDoesNotExist(Exception):
    pass


class UniqueConstraintError(Exception):
    pass



from collections import defaultdict
from collections.abc import Iterable
from typing import Optional, Generator, Union, Hashable, List, Dict
import sys

import dataset

from pymemdb import Column, ColumnDoesNotExist


version = sys.version_info
if version.major < 3:  # pragma: no cover
    raise Exception("Python version must be 3.6 or higher")
if version.major >= 3 and version.minor < 8:  # pragma: no cover
    ORDER_TYPE = Union[str, bool]
else:  # pragma: no cover
    from typing import Literal  # type: ignore
    ORDER_TYPE = Literal["ascending", "descending", False]  # type:ignore

ROW_GEN = Generator[dict, None, None]


class Table:
    """Object that represents a Table in the Database.
       Can also used standalone"""

    def __init__(self, name: Optional[str] = None,
                 primary_id: str = "id") -> None:
        self.name = name
        self.idx_name = primary_id
        self._columns: defaultdict = defaultdict(Column)
        self.idx = 0
        self.keys: set = set()
        self.create_column(name=self.idx_name, unique=True)

    @classmethod
    def from_dataset(cls, table: dataset.table.Table, name=None) -> None:
        if name is None:
            name = table.name

        pk = None
        for col in table.table.columns:
            if col.primary_key is True:
                pk = col.name
                break

        pymemdb_table = cls(name, primary_id=pk)
        for row in table:
            pymemdb_table.insert(row)

        return pymemdb_table

    def to_dataset(self, db: dataset.Database, drop=False):
        if self.name in db.tables and drop:
            db[self.name].drop()
        if self.name not in db.tables:
            db.create_table(self.name, primary_id=self.idx_name)

        with db as tx:
            for row in self.all():
                tx[self.name].insert(row)

    def all(self, ordered: ORDER_TYPE = False) -> ROW_GEN:
        """returns a generator of all rows of the table.

        Keyword Arguments:
            ordered {["ascending", "descending", False]} -- Can be set to
                ascending or descending to sort by the primary id column
                (default: {False})
        Raises:
            ValueError: [if 'ordered' is not in {"ascending', 'descending',
                                                 False}]

        Returns:
            Generator -- [A Generator over all the rows in the table]

        Yields:
            dict -- [Dictionary that contains all elements of a row in the
                     table]
        """
        if ordered is False:
            for i in self.keys:
                yield self._get_row(i)
        elif ordered == "ascending":
            for i in sorted(self.keys):
                yield self._get_row(i)
        elif ordered == "descending":
            for i in sorted(self.keys, reverse=True):
                yield self._get_row(i)
        else:
            raise ValueError("Value for kwarg 'ordered' not in [False, "
                             "ascending, descending] !")

    def create_column(self, name: str, default: Hashable = None,
                      unique: bool = False) -> None:
        """Create a Column in the table.

        Arguments:
            name {str} -- Name of the column

        Keyword Arguments:
            default {Hashable} -- Default value for the column
                                    (default: {None})
            unique {bool} -- If values are enforced to be unique for this
                             column. If True, trying to insert a value more
                             than once will raise UniqueConstraintError
                             (default: {False})
        """
        self._columns[name] = Column(default=default, unique=unique)

    @property
    def columns(self) -> List[str]:
        """Returns a list of all column names of the table.

        Returns:
            List[str] -- [List of column names]

        """
        return list(self._columns)

    def drop(self) -> None:
        """Delete table."""
        del self

    def insert(self, row: Dict) -> int:
        """Inserts a row in the table. If a column is not present,
           it will be created with default value None

        Arguments:
            row {Dict} -- dictionary that represents a row in the table.

        Raises:
            UniqueConstraintError: [if constraint of a column is violated]

        Returns:
            int -- [primary key for the row inserted]
        """
        if self.idx_name in row:
            idx = row[self.idx_name]
        else:
            while self.idx in self.keys:
                self.idx += 1
            idx = self.idx
        self.keys.add(idx)
        for key, val in row.items():
            self._columns[key].insert(idx, val)
        if self.idx_name not in row:
            self._columns[self.idx_name].insert(idx, idx)
        return idx

    def insert_ignore(self, row: Dict, keys: List[str], ignore_errors: bool = True) -> Optional[int]:
        """Inserts rows into the table. If another row is already present
           where all the values are identical for the fields in 'keys', the
           insert will be skipped

        Arguments:
            row {Dict} -- [row to be inserted]
            keys {List[str]} -- [List of columns to check for ignore]

        Raises:
            UniqueConstraintError: [if constraint of a column is violated]

        Returns:
            Optional[int] -- [primary key for the row inserted or None if
                              the insert was skipped]

        """
        results = self.find(**{key: row[key] for key in keys}, ignore_errors=ignore_errors)
        try:
            next(results)
        except StopIteration:
            return self.insert(row)
        return None

    def find(self, ignore_errors: bool = True,
             **kwargs) -> Generator[dict, None, None]:
        """finds rows in the table.

        Keyword Arguments:
            **kwargs -- keyword is the column name and value represents the
                        value to search.
                        if value is in iterable, it matches a
                        SELECT * WHERE keyword IN val
                        search.
            ignore_errors {bool} -- if True, it raises an error if a column
                                    does not exist in the table
                                    (default: {False})

        Returns:
            Generator[dict] -- [Generator over all rows that match the search]
        """
        results = self._find_rows(ignore_errors=ignore_errors, **kwargs)

        for idx in results:
            yield {self.idx_name: idx, **self._get_row(idx)}

    def find_one(self, ignore_errors: bool = False, **kwargs) -> Optional[dict]:
        """finds a single row from the table, if there is one.
        Keyword Arguments:
            ignore_errors {bool} -- if True, it raises an error if a column
                                    does not exist in the table
                                    (default: {False})

        Returns:
            Optional[dict] -- single row from the table or None if there is
                              none that matches the query.
        """

        try:
            row = next(self.find(ignore_errors=ignore_errors, **kwargs))
        except StopIteration:
            return None
        return row

    def delete(self, ignore_errors: bool = False, **kwargs) -> int:
        pks = {row[self.idx_name] for row in self.find(**kwargs)}
        if len(pks) == 0 and not ignore_errors:
            raise KeyError(f"No matching rows found for {kwargs}")
        for pk in pks:
            self.keys.remove(pk)
        for col in self._columns.values():
            for pk in pks:
                col.drop(pk)
        return len(pks)

    def update(self, where: dict, **kwargs) -> int:
        pks = self._find_rows(**where)
        if not pks:
            return 0

        for col, val in kwargs.items():
            cell_dict = self._columns[col].cells
            val_dict = self._columns[col].values
            for pk in pks:
                cell_dict[pk] = val
                val_dict[val].add(pk)
        return len(pks)

    def update_replace(self, where: dict, **kwargs):
        n_rows = self.update(where=where, **kwargs)
        if n_rows == 0:
            return 0

        new_where = {**where, **kwargs}
        rows = self.find(**new_where)
        result_counter = defaultdict(frozenset)
        for row in rows:
            pk = row[self.idx_name]
            del row[self.idx_name]
            r = result_counter[tuple(row.items())]
            result_counter[tuple(row.items())] = r.union({pk})

        rowcount = 0
        for pks in result_counter.values():
            if len(pks) < 2:
                continue
            keep = min(pks)
            pks = pks.difference({keep})
            n = self.delete(**{self.idx_name: pks})
            rowcount += n

        return rowcount

    def _get_row(self, idx: int) -> dict:
        row = {col: self._columns[col].find_value(idx) for col in self.columns}
        row = {self.idx_name: idx, **row}
        return row

    def _find(self, col: str, val: Hashable) -> set:
        if isinstance(val, Iterable) and not isinstance(val, str):
            results: set = set()
            for v in val:
                results.update(self._columns[col].find(v))
            return results
        return self._columns[col].find(val)

    def _find_rows(self, ignore_errors: bool = True, **kwargs) -> set:
        results: set = set()
        for col, val in kwargs.items():
            if col not in self._columns:
                if ignore_errors:
                    continue
                else:
                    raise KeyError(f"Column {col} not in Table!")
            if len(results) == 0:
                results = self._find(col, val)
            else:
                pk = self._find(col, val)
                results = results.intersection(pk)
            if val == self._columns[col].default:
                column_cells = set(self._columns[col].cells)
                mis_def_keys = self.keys.symmetric_difference(column_cells)
                results.update(mis_def_keys)
            if not results:
                return set()
        return results

    def __eq__(self, other):
        return self.name == other.name

    def __getitem__(self, col):
        if col not in self._columns:
            raise ColumnDoesNotExist("Column {col} does not exist!")
        return self._columns[col]

    def __delitem__(self, col):
        if col not in self._columns:
            raise ColumnDoesNotExist(f"Column {col} does not exist!")
        del self._columns[col]

    def __len__(self):
        return len(self.keys)



import dataset

import pymemdb


def test_simple_db():
    db = dataset.Database("sqlite:///:memory:")

    table1 = db.create_table("table1", primary_id="a")
    table1.insert(dict(a=1, b="2"))

    table2 = db.create_table("table2")
    table2.insert(dict(name="John", lastname="Doe"))

    assert db.tables == ["table1", "table2"]
    pymem_db = pymemdb.Database.from_dataset(db)

    assert list(pymem_db["table1"].all()) == list(table1.all())
    assert list(pymem_db["table2"].all()) == list(table2.all())



"""A very fast in-memory database with export to sqlite written purely in python"""

from .errors import *
from .column import Column
from .table import Table
from .database import Database


__version__ = "1.4.2"



from pymemdb import Table

def test_single_insert_auto_primary():
    table = Table(primary_id="pk")
    row = {"a": 1, "b": 2, "c": 0}
    table.insert(row)
    assert list(table.all()) == [{"pk": 0, **row}]


def test_single_insert_given_primary():
    table = Table(primary_id="pk")
    row = {"pk": 1, "a": 1, "b": 2, "c": 0}
    table.insert(row)
    assert list(table.all()) == [{**row}]



def test_no_insert_columns_on_find():
    t = Table(primary_id="a")
    t.insert({"a": 1})
    t.insert({"a": 2})

    assert list(t.find(a=1, b=2, ignore_errors=True)) == [{"a": 1}]
    assert t.columns == ["a"]



import os
from setuptools import setup, find_packages

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ..


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="pymemdb",
    version="1.4.2",
    author_email="luerhard@googlemail.com",
    description=("A very fast in-memory database with export to sqlite written purely in python"),
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    keywords="database rdbms sqlite relational python dictionary fast build memory",
    license="MIT",
    url="https://github.com/luerhard/pymemdb",
    install_requires=[
        "dataset",
    ],
    packages=find_packages(exclude=["tests/",
                                    ".circleci/",
                                    ]),
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities"
    ],
)



from pymemdb import Table


def test_update_single():
    t = Table()

    row1 = dict(vorname="rainer", nachname="greiff")
    row2 = dict(vorname="rainer", nachname="greiff")
    row3 = dict(vorname="rainer", nachname="erhard")

    for row in [row1, row2, row3]:
        t.insert(row)

    row_vals_before = set(t["vorname"].values["rainer"])
    t.update(dict(vorname="rainer", nachname="erhard"), n_values=45)

    row_vals_after = set(t["vorname"].values["rainer"])
    result = list(t.find(n_values=45))
    assert len(result) == 1
    assert row_vals_before == row_vals_after


def test_update_replace():
    t = Table()

    row1 = dict(vorname="rainer", nachname="greiff")
    row2 = dict(vorname="rainer", nachname="smith")
    row3 = dict(vorname="rainer", nachname="erhard")

    for row in [row1, row2, row3]:
        t.insert(row)

    t.update_replace(where={"nachname": "smith"}, nachname="greiff")

    assert len(t) == 2
    assert len(list(t.find(nachname="greiff"))) == 1


def test_update_replace_with_context():
    t = Table()

    row1 = dict(vorname="rainer", nachname="greiff", address="123street")
    row2 = dict(vorname="rainer", nachname="smith", address="123street")
    row3 = dict(vorname="rainer", nachname="smith", address="456street")
    row4 = dict(vorname="luke", nachname="doe", address="23Street")

    for row in [row1, row2, row3, row4]:
        t.insert(row)

    t.update_replace(where={"nachname": "smith"}, nachname="greiff")

    assert len(t) == 3
    assert len(list(t.find(nachname="greiff"))) == 2


def test_update_replace_no_update():
    t = Table()

    row1 = dict(vorname="rainer", nachname="greiff", address="123street")
    row2 = dict(vorname="rainer", nachname="smith", address="123street")
    row3 = dict(vorname="rainer", nachname="smith", address="456street")
    row4 = dict(vorname="luke", nachname="doe", address="23Street")

    for row in [row1, row2, row3, row4]:
        t.insert(row)

    t.update_replace(where={"nachname": "pete"}, nachname="greiff")

    assert len(t) == 4
    assert len(list(t.find(nachname="greiff"))) == 1


def test_update_replace_no_replace():
    t = Table()

    row1 = dict(vorname="rainer", nachname="greiff", address="123street")
    row2 = dict(vorname="rainer", nachname="smith", address="123street")
    row3 = dict(vorname="rainer", nachname="smith", address="456street")
    row4 = dict(vorname="luke", nachname="doe", address="23Street")

    for row in [row1, row2, row3, row4]:
        t.insert(row)

    t.update_replace(where={"nachname": "doe"}, nachname="greiff")

    assert len(t) == 4
    assert len(list(t.find(nachname="greiff"))) == 2


def test_update_replace_multiple_replace():
    t = Table()

    rows = [
        dict(vorname="rainer", nachname="greiff", address="123street"),
        dict(vorname="rainer", nachname="greiff", address="123street"),
        dict(vorname="rainer", nachname="smith", address="123street"),
        dict(vorname="rainer", nachname="smith", address="123street"),
        dict(vorname="rainer", nachname="smith", address="123street"),
        dict(vorname="rainer", nachname="smith", address="456street"),
        dict(vorname="luke", nachname="doe", address="23Street"),
    ]

    for row in rows:
        t.insert(row)

    t.update_replace(where={"nachname": "smith"}, nachname="greiff")

    assert len(t) == 3
    assert len(list(t.find(nachname="greiff"))) == 2
